// CaptureButton.cpp : implementation file
//

#include "stdafx.h"
#include "ScreenCapture.h"
#include "CaptureButton.h"
#include "mousehook.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

static UINT UWM_HOOK = ::RegisterWindowMessage(UWM_HOOK_MSG);


#define IDT_BLINK 1
/////////////////////////////////////////////////////////////////////////////
// CCaptureButton

CCaptureButton::CCaptureButton()
   {
    blinking = FALSE;
    blink = FALSE;
    cancelling = FALSE;
    widepen.CreatePen(PS_SOLID, 2 * ::GetSystemMetrics(SM_CXBORDER), RGB(0,0,0));
    fillcolor = RGB(255,255,255);
    menugrab = FALSE;
   }

CCaptureButton::~CCaptureButton()
{
}


BEGIN_MESSAGE_MAP(CCaptureButton, CButton)
        ON_REGISTERED_MESSAGE(UWM_HOOK, OnHookMessage)
	//{{AFX_MSG_MAP(CCaptureButton)
	ON_WM_TIMER()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_SETCURSOR()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CCaptureButton message handlers

/****************************************************************************
*                          CCaptureButton::DrawItem
* Inputs:
*       LPDRAWITEMSTRUCT dis: Draw data
* Result: void
*       
* Effect: 
*       Draws the button
* Notes:
*	This is actually a general button-drawing subroutine that has been
*	cloned into this handler, and is therefore somewhat more general
*	than is strictly required.
****************************************************************************/

void CCaptureButton::DrawItem(LPDRAWITEMSTRUCT dis) 
   {
    CDC * dc = CDC::FromHandle(dis->hDC);
    CRect r(dis->rcItem);
    int saved = dc->SaveDC();
    CPoint pt;
    CSize sz;
    CSize border(::GetSystemMetrics(SM_CXBORDER), ::GetSystemMetrics(SM_CYBORDER));
    CPoint baseOffset(border.cx, border.cy);

    dc->SetTextColor(::GetSysColor(COLOR_BTNTEXT));
    dc->SetBkMode(TRANSPARENT);
    if(dis->itemState & (ODS_DISABLED | ODS_GRAYED))
       { /* grayed */
	dc->SetTextColor(::GetSysColor(COLOR_GRAYTEXT));
       } /* grayed */

    DWORD style = GetStyle();
    style &= (BS_LEFT | BS_RIGHT | BS_CENTER);

#define MARGIN 4
    CString s;
    GetWindowText(s);
    sz = dc->GetTextExtent(s);

    UINT flags = (GetStyle() & BS_MULTILINE ? 0 : DT_SINGLELINE);
    switch(style)
       { /* hstyle */
	case BS_LEFT:
	   pt.x = MARGIN * border.cx;
	   flags |= DT_LEFT;
	   break;
	case BS_RIGHT:
	   pt.x = r.Width() - MARGIN * border.cx - sz.cx;
	   flags |= DT_RIGHT;
	   break;
	case 0:
	case BS_CENTER:
	   pt.x = (r.Width() - sz.cx) / 2;
	   flags |= DT_CENTER;
	   break;
       } /* hstyle */

    style = GetStyle();
    style &= (BS_TOP | BS_BOTTOM | BS_VCENTER);

    switch(style)
       { /* vstyle */
	case BS_TOP:
	   pt.y = 3 * border.cy;
	   flags |= DT_TOP;
	   break;
	case BS_BOTTOM:
	   pt.y = r.Height() - MARGIN * border.cy - sz.cy;
	   flags |= DT_BOTTOM;
	   break;
	case 0:
	case BS_VCENTER:
	   pt.y = (r.Height() - sz.cy) / 2;
	   flags |= DT_VCENTER;
	   break;
       } /* vstyle */

     // Compute the focus rectangle area
    CRect focus(pt.x, pt.y, pt.x + sz.cx, pt.y + sz.cy);

     // For visual effect, if the button is down, shift the text over and
     // down to make it look "pushed".
    CPoint useOffset(0, 0);
    if(blinking)
       { /* down */
	useOffset = baseOffset;
       } /* down */

    DWORD edge = BF_RECT | BF_SOFT;

    // "Blinking" tells us if the button should be drawn
    // down or up. I have hardwired red and green colors
    // for the blinking
    if(blinking)
       { /* blinking */
	if(blink)
	   dc->FillSolidRect(&r, RGB(255, 0, 0));
	else
	   dc->FillSolidRect(&r, RGB(0, 255, 0));
       } /* blinking */
    else
       { /* no blink */
	edge |= BF_MIDDLE;
       } /* no blink */

     // Draw the traditional pushbutton edge using DrawEdge
    if(blinking)
       { /* down */
	dc->DrawEdge(&dis->rcItem, EDGE_SUNKEN, BF_RECT | BF_SOFT);
       } /* down */
    else
       { /* up */
	dc->DrawEdge(&dis->rcItem, EDGE_RAISED, BF_RECT | BF_SOFT);
       } /* up */

     // Adjust the focus position and text position by shifting them
     // right-and-down by the desired offset
    focus += useOffset;
    pt += useOffset;

    focus.InflateRect(border.cx, border.cy);

    r.InflateRect(-MARGIN * border.cx, -MARGIN * border.cy);
    r += useOffset;

    dc->DrawText(s, r, flags);

    if(dis->itemState & ODS_FOCUS)
       ::DrawFocusRect(dis->hDC, &focus);

    dc->RestoreDC(saved);
   }

/****************************************************************************
*                           CCaptureButton::OnTimer
* Inputs:
*       UINT nIDEvent: ignored because we have only one timer event
* Result: void
*       
* Effect: 
*       Causes the button color to shift by complementing the blink state
*	and forcing a redraw
****************************************************************************/

void CCaptureButton::OnTimer(UINT nIDEvent) 
   {
    blink = !blink;
	
    InvalidateRect(NULL);
    CButton::OnTimer(nIDEvent);
   }

/****************************************************************************
*                          CCaptureButton::setButton
* Result: void
*       
* Effect: 
*       Sets the button blink mode
****************************************************************************/

void CCaptureButton::setButton()
    {
     // Enable the blinking button effect
     blinking = TRUE;
     blink = TRUE;
     SetTimer(IDT_BLINK, 350, NULL);

     // Save the window caption string so we can
     // reset it in OnLButtonUp, since the
     // OnMouseMove will replace it with the
     // window handle
     GetWindowText(caption);
    } // CCaptureButton::setButton

/****************************************************************************
*                        CCaptureButton::OnLButtonDown
* Inputs:
*       UINT nFlags: flags, ignored except to pass to superclass
*	CPoint point: Point of button down click, in client coordinates
* Result: void
*       
* Effect: 
*       Initializes the capture button state
****************************************************************************/

void CCaptureButton::OnLButtonDown(UINT nFlags, CPoint point) 
   {
    // Set the capture and initialize the variables
    // used by OnMouseMove and OnLButtonUp
    lastCapture = NULL;
    if(!menugrab)
       { /* non-menu grab */
	SetCapture();
	cursor = NULL;
	setButton();
       } /* non-menu grab */
    else
       { /* enable/disable hook */
	if(!blinking)
	   { /* set mode */
	    if(!setHook(m_hWnd))
	       { /* failed */
		DWORD err = ::GetLastError();
		ASSERT(FALSE);
	       } /* failed */
	    setButton();
	   } /* set mode */
	else
	   { /* clear it */
	    cancelling = TRUE;
	    if(!clearHook(m_hWnd))
	       { /* failed to clear */
		DWORD err = ::GetLastError();
		ASSERT(FALSE);
	       } /* failed to clear */
	    KillTimer(IDT_BLINK);
	   } /* clear it */
       } /* enable/disable hook */

    CButton::OnLButtonDown(nFlags, point);
}

/****************************************************************************
*                        CCaptureButton::cancelButton
* Result: void
*       
* Effect: 
*       Cancels the button state
****************************************************************************/

void CCaptureButton::cancelButton()
    {
     // This code turns off the blinking-button effect
     blinking = FALSE;
     blink = FALSE;
     KillTimer(IDT_BLINK);

     // Force the button to redraw
     InvalidateRect(NULL);

     // Now restore the caption of the button to its original state
     CButton::SetState(FALSE);
     SetWindowText(caption);
    } // CCaptureButton::cancelButton

/****************************************************************************
*                         CCaptureButton::GrabWindow
* Result: void
*       
* Effect: 
*       Captures the window image
****************************************************************************/

void CCaptureButton::GrabWindow(CPoint point)
    {
     // Point is relative to this window. We need to convert it to
     // client coordinates of the parent in order to locate the
     // window to display
     CPoint mappt(point);
     MapWindowPoints(GetParent(), &mappt, 1);

     // Get the child window (if there is no child window under the
     // cursor, the parent window is returned)
     CWnd * wnd = GetParent()->ChildWindowFromPoint(mappt);
     if(wnd == NULL)
	{ /* see if other window */
	 CPoint screen(point);
	 ClientToScreen(&screen);
	 wnd = CWnd::WindowFromPoint(screen);
	} /* see if other window */
     if(wnd != NULL)
	{ /* capture this window */
	 // There is a window. We must make sure it is not our own
	 // blinking button window
	 if(wnd != this)
	    { /* can capture */
	     // We had drawn the highlight inside the client area
	     // of the target window; to get a clean capture, we
	     // have to force the window to be redrawn
	     wnd->InvalidateRect(NULL);
	     wnd->UpdateWindow();

	     // Get a DC for the entire window under the cursor
	     CDC dc;
	     HDC hdc = ::GetWindowDC(wnd->m_hWnd);
	     dc.Attach(hdc);

	     // Create a memory DC into which the bitmap will be captured
	     CDC memDC;
	     memDC.CreateCompatibleDC(&dc);

	     // If there is already a bitmap, delete it as we are going to
	     // replace it
	     bmp.DeleteObject();

	     // Compute the size of the bitmap we are going to need
	     CRect r;
	     wnd->GetWindowRect(&r);

	     CPoint offset(0, 0);
	     CSize sz(r.Width(), r.Height());
	     // We may need to expand the width and height to allow the cursor to be
	     // drawn, because it may overhang the edges of the window
	     // Note that no matter how large the cursor *actually* is, we have
	     // to treat it as a full 32x32 cursor (the logic required to determine
	     // its actual size is somewhat complex, and involves getting the bitmap
	     // for the cursor's AND-mask and computing a bounding box. This is more
	     // trouble than I deem it to be worth)
	     //
	     // We need to compute the top left corner and bottom right corner of
	     // the cursor. Note the hotspot must be within the bounds of the window
	     // or it wouldn't be in this window.
	     //
	     //     +=============================================+
	     //     |                                             |
	     // tlc |                                             |
	     //  +--|--+                                       +--|--+
	     //  |  | a|                                       |b |  |
	     //  |  |  |                                       |  |  |
	     //  +--|--+                                       +--|--+
	     //     |                                             |  brc
	     //     +=============================================+
	     //  hotspot: a = (5, 1)
	     //  hotspot: b = (1, 1)
	     //  Converted to client coordinates:
	     //  Cursor is at x,y
	     //  a = x,y
	     //      tlc = x - a.x, y - a.y
	     //  b = x,y
	     //      brc = x - b.x + ::GetSystemMetrics(SM_CXCURSOR)
	     //            y - b.y + ::GetSystemMetrics(SM_CYCURSOR)

	     // The ICONINFO structure applies to both cursors and
	     // icons and allows us to find out where the hotspot is
	     ICONINFO info;

	     // We need to remap the point to client coordinates in
	     // the target window
	     CPoint wndpt(point);
	     MapWindowPoints(wnd, &wndpt, 1);

	     if(GetIconInfo((HICON)::GetCursor(), &info))
		{ /* got info */
		 // Compute the top-left-corner tlc of the cursor
		 CPoint tlc(wndpt.x - info.xHotspot, wndpt.y - info.yHotspot);

		 // If the cursor overhangs on the left or top, we need
		 // to change the offset at which we will copy the window
		 // image into the target bitmap so there is room to show the
		 // cursor
		 if(tlc.x < 0)
		    { /* hangs off left */
		     int dx = info.xHotspot - wndpt.x;
		     offset.x += dx;
		     sz.cx += dx;
		    } /* hangs off left */
		 if(tlc.y < 0)
		    { /* hangs off top */
		     int dy = info.yHotspot - wndpt.y;
		     offset.y += dy;
		     sz.cy += dy;
		    } /* hangs off top */

		 // If the cursor overhangs on the right or bottom, we
		 // need to extend the bitmap so there is room to
		 // display the cursor beyond the edges

		 // To do this, we need to compute the bottom right
		 // corner of the cursor, which is always
		 // SM_CXCURSOR,SM_CYCURSOR from the tlc
		 CPoint brc(tlc.x + ::GetSystemMetrics(SM_CXCURSOR),
			    tlc.y + ::GetSystemMetrics(SM_CYCURSOR));
		 if(brc.x > sz.cx)
		    { /* hangs off right */
		     sz.cx = brc.x;
		    } /* hangs off right */
		 if(brc.y > sz.cy)
		    { /* hangs off bottom */
		     sz.cy = brc.y;
		    } /* hangs off bottom */
		} /* got info */

	     //----------------------------------------------------------
	     //     +---------------------------------------+-+
	     //     | Caption            a                  |X|
	     //     +---------------------------------------+-+
	     //     | Menu                                    |
	     //     +-----------------------------------------+
	     //     |                                         |
	     //     |                    b                    |
	     //     |                                         |
	     //     |                                         |  overlapped window
	     //     +-----------------------------------------+
	     //
	     //          +-----------+
	     //          |     c     |   Child window
	     //          +-----------+
	     //
	     //         
	     // Now compute the difference between the client area and the
	     // window area so we have coordinates properly adjusted
	     CPoint cursorOffset;
	     CRect c;
	     wnd->GetClientRect(&c);
	     UINT style = wnd->GetStyle();
	     
	     if( (style & (WS_OVERLAPPED | WS_POPUP | WS_CHILD)) == WS_OVERLAPPED)
		{ /* overlapped window */
		 CRect w(r);
		 GetParent()->ScreenToClient(&w);
		 //CPoint cursorOffset((r.Width() - c.Width()) / 2,
		 //		    (r.Height() - c.Height()) / 2);
		 //CPoint cursorOffset(0,0);
		 CPoint cursorOffset((r.Width() - c.Width()) / 2,
				     c.top - w.top); 
		} /* overlapped window */
	     else
		{ /* child window */
		 cursorOffset = CPoint( (r.Width() - c.Width()) / 2,
					(r.Height() - c.Height() ) / 2);
		} /* child window */
	     // Now create the bitmap into which we are going to
	     // capture the image
	     bmp.CreateCompatibleBitmap(&dc, sz.cx, sz.cy);
	     CBitmap * oldbm = memDC.SelectObject(&bmp);

	     // Before we copy the image in, we blank the bitmap to
	     // the background fill color
	     memDC.FillSolidRect(&CRect(0,0,sz.cx, sz.cy), fillcolor);

	     // Copy the window image from the window DC into the memory DC
	     // We use the offset previously computed to allow space for
	     // the cursor to be displayed, if necessary
	     // Note that because the cursor is always treated as 32x32, we
	     // may have cases where there is extra space around the window
	     // which appears to be unnecessary. The only option is to
	     // compute the bounding box of the actual cursor image
	     memDC.BitBlt(offset.x, offset.y, r.Width() + 1, r.Height() + 1, &dc, 0, 0, SRCCOPY);

	     // If the window has a caption bar, we need to offset the y coordinate
	     // to account for this (no, I don't know why)
	     if((wnd->GetStyle() & WS_CAPTION) == WS_CAPTION)
		{ /* has caption */
		 offset.y += ::GetSystemMetrics(SM_CYCAPTION);
		} /* has caption */
	     // Now draw the image of the cursor that we captured during
	     // the mouse move. DrawIcon will draw a cursor as well.
	     memDC.DrawIcon(wndpt.x - info.xHotspot + offset.x + cursorOffset.x,
			    wndpt.y - info.yHotspot + offset.y + cursorOffset.y,
			    (HICON)cursor);

	     // Now deselect the bitmap from the memory DC. 
	     memDC.SelectObject(oldbm);

	    } /* can capture */
	 else
	    { /* can't capture */
	     // I could have chosen to leave the bitmap alone;
	     // I chose to delete it if there was some error
	     bmp.DeleteObject();
	     SetWindowText(_T(""));
	     ::SetCursor(AfxGetApp()->LoadStandardCursor(IDC_NO));
	    } /* can't capture */
	} /* capture this window */
     else
	{ /* no child window */
	    // If there was no child window found, delete the bitmap
	 bmp.DeleteObject();
	} /* no child window */

     cancelButton();
    } // CCaptureButton::GrabWindow

/****************************************************************************
*                         CCaptureButton::OnLButtonUp
* Inputs:
*       UINT nFlags: ignored
*	CPoint point: Point, in client coordinates
* Result: void
*       
* Effect: 
*       Captures the bitmap for the child window under the cursor
****************************************************************************/

void CCaptureButton::OnLButtonUp(UINT nFlags, CPoint point) 
   {
    if(menugrab && blinking)
       { /* menu grabber */
        if(cancelling)
            cancelButton();
	cancelling = FALSE;
	CButton::OnLButtonUp(nFlags, point);
	ReleaseCapture();
	return;
       } /* menu grabber */
    
    if(blinking && GetCapture() == this)
       { /* had capture */
	GrabWindow(point);

	// We have finished processing; release the mouse
	ReleaseCapture();
       } /* had capture */


    // We need to simulate the OnLButtonUp operation
    // Which, due to the defective design of MFC, we cannot simulate
    // by passing an updated point!
    // This is because MFC stupidly uses the values of the last
    // message to process in the base class, an inexcusable
    // piece of "optimization" that makes the system work incorrectly
    // and inconsistently.
    GetParent()->SendMessage(WM_COMMAND, (WPARAM)MAKELONG(GetDlgCtrlID(), BN_CLICKED),
			                 (LPARAM)m_hWnd);
    // CButton::OnLButtonUp(nFlags, point);
}

/****************************************************************************
*                          CCaptureButton::Highlight
* Inputs:
*       CWnd * wnd: Window reference
* Result: void
*       
* Effect: 
*       Highlights the window. Unhighlights the previous window
****************************************************************************/

void CCaptureButton::Highlight(CWnd * wnd, CPoint point)
    {
     // We have a child window. If it is the same child
     // window we had before, we don't need to do anything
     // special; otherwise we want to highlight the window
     if(wnd != lastCapture)
	{ /* changed */
	 // If we had a previous capture, we need to force
	 // that window to redraw properly to get rid of the
	 // highlighting we have done
	 if(lastCapture != NULL)
	    { /* redraw old */
	     lastCapture->InvalidateRect(NULL);
	     lastCapture->UpdateWindow();
	    } /* redraw old */

	 // Compute the dimensions of the client area of
	 // the window so we know how large to create the
	 // highlighting area
	 CRect r;
	 wnd->GetClientRect(&r);

	 // Create a DC so we can draw the highlight
	 // We want to draw a hollow rectangle with thick
	 // sides. The 'widepen' defines the width of the
	 // highlighting rectangle.

	 CClientDC dc(wnd);
	 dc.SelectStockObject(NULL_BRUSH);
	 dc.SelectObject(&widepen);
	 dc.Rectangle(&r);

	 // Change the caption of the window to be the
	 // window handle
	 CString s;
	 s.Format(_T("%08x"), wnd->m_hWnd);
	 SetWindowText(s);
	 lastCapture = wnd;
	} /* changed */

     // Ask the window what cursor it should have by
     // sending a WM_SETCURSOR message. To do this, it needs to
     // know the "hit code" so it must first query the window
     // using WM_NCHITTEST
     // Note that we special-case HTNOWHERE to convert it to
     // HTCLIENT (for some reason this is required to get the
     // cursor to be an I-beam in an edit control)
     CPoint screen(point);
     ClientToScreen(&screen);
     UINT hit = wnd->SendMessage(WM_NCHITTEST, 0, MAKELPARAM(screen.x, screen.y));
     if(hit == HTNOWHERE)
	hit = HTCLIENT;
     if(hit == HTMENU)
	{ /* in menu bar */
	 if(!inmenu)
	    wnd->SendMessage(WM_NCLBUTTONDOWN, HTMENU, MAKELPARAM(screen.x, screen.y));
	 inmenu = TRUE;
	} /* in menu bar */
     wnd->SendMessage(WM_SETCURSOR, (WPARAM)(wnd->m_hWnd), MAKELPARAM(hit, WM_MOUSEMOVE));
     // The side effect of sending WM_SETCURSOR is that the cursor 
     // should now be the correct cursor for the window, as determined
     // either by its OnSetCursor handler or by DefWindowProc
     // setting the class cursor. The cursor could be NULL.
     cursor = ::GetCursor();
    } // CCaptureButton::Highlight

/****************************************************************************
*                         CCaptureButton::OnMouseMove
* Inputs:
*       UINT nFlags: ignored except to pass to superclasss
*	CPoint point: The client coordinates of the mouse, in the
*		coordinate system of this control
* Result: void
*       
* Effect: 
*       Determines the windown under the point, and forces the
*	cursor for that window to be computed.
****************************************************************************/

void CCaptureButton::OnMouseMove(UINT nFlags, CPoint point) 
   {
    // If the mouse moves over the window and we don't have
    // capture, just handle it normally
    CPoint parent(point);
    CWnd * captured = GetCapture();
    if(blinking && captured == this)
       { /* has capture */
	// Convert the point to coordinates in the parent
        MapWindowPoints(GetParent(), &parent, 1);

	// Obtain a reference to the child window under the
	// mouse
	CWnd * wnd = GetParent()->ChildWindowFromPoint(parent);
	if(wnd != NULL)
	   { /* has window */
	    Highlight(wnd, point);
	   } /* has window */
	else
	   { /* no window */
	    CPoint screen(point);
	    ClientToScreen(&screen);
	    CWnd * wnd = CWnd::WindowFromPoint(screen);
	    if(wnd != NULL)
	       Highlight(wnd, point);
	   } /* no window */
	
       } /* has capture */
    
    CButton::OnMouseMove(nFlags, point);
   }

/****************************************************************************
*                         CCaptureButton::OnSetCursor
* Inputs:
*       CWnd * pWnd: Window reference
*	UINT nHitTest: ignored
*	UINT message: ignored
* Result: BOOL
*       TRUE if cursor has been set
* Effect: 
*       If the cursor is over the current button, sets the "No" cursor
*	and returns TRUE
****************************************************************************/

BOOL CCaptureButton::OnSetCursor(CWnd* pWnd, UINT nHitTest, UINT message) 
   {
    if(pWnd == this && GetCapture() == this)
       { /* set no cursor */
	::SetCursor(AfxGetApp()->LoadStandardCursor(IDC_NO));
	return TRUE;
       } /* set no cursor */
	
    return CButton::OnSetCursor(pWnd, nHitTest, message);
   }

/****************************************************************************
*                        CCaptureButton::OnHookMessage
* Inputs:
*       WPARAM: Message ID
*	LPARAM: unused
* Result: LRESULT
*       Logically void, 0, always
* Effect: 
*       Handles message dropdowns
****************************************************************************/

LRESULT CCaptureButton::OnHookMessage(WPARAM wParam, LPARAM)
    {
     GetParent()->SendMessage(UWM_HOOK, wParam, 0);
     CPoint point;
     switch(wParam)
	{ /* wParam */
	 case WM_MENUSELECT:
	    break;
	 case WM_ENTERMENULOOP:
	    break;
	 case WM_EXITMENULOOP:
	    clearHook(m_hWnd);
	    cancelButton();
	    GetCursorPos(&point);
	    ScreenToClient(&point);
	    GrabWindow(point);
	    break;
	} /* wParam */
     return 0;
    } // CCaptureButton::OnHookMessage
